import { DIFF_DELETE, DIFF_INSERT, DIFF_EQUAL, diff_match_patch } from 'diff-match-patch';
import { createOptions as createOptions$1, scan } from '@emmetio/html-matcher';
import { isNumber } from '@emmetio/scanner';

const defaultOptions = {
    compact: true,
    normalizeSpace: true,
    allTokens: true,
    baseStart: 0,
    preserveTags: [],
    inlineElements: [
        'a', 'abbr', 'acronym', 'applet', 'b', 'basefont', 'bdo',
        'big', 'br', 'button', 'cite', 'code', 'del', 'dfn', 'em', 'font', 'i',
        'iframe', 'img', 'input', 'ins', 'kbd', 'label', 'map', 'object', 'q',
        's', 'samp', 'select', 'small', 'span', 'strike', 'strong', 'sub', 'sup',
        'textarea', 'tt', 'u', 'var'
    ]
};
function createOptions(opt) {
    return createOptions$1(Object.assign(Object.assign({}, defaultOptions), opt));
}

const wordDelimiters = '.,…/\\!?:;()[]{}<>"\'«»“”‘’-–—\n\r\t \u00a0'.split('').map(ch => ch.charCodeAt(0));
const numberPunct = '.,'.split('').map(ch => ch.charCodeAt(0));
/**
 * Returns updated list of diff ops so that adjacent delete and insert ops include
 * word bounds
 */
function updateWordBounds(diffs) {
    diffs = [...diffs];
    const result = [];
    let i = 0;
    let start = null;
    while (i < diffs.length) {
        const cur = diffs[i];
        const next = diffs[i + 1];
        if (cur[0] === DIFF_DELETE && next && next[0] === DIFF_INSERT) {
            const del = [...cur];
            const ins = [...next];
            if (start) {
                // Move start word bound, if required
                moveStartBound(start, result, del, ins);
            }
            // Advance next until we find non-updated chunk with valid word bound
            i += 2;
            while (i < diffs.length) {
                const peek = diffs[i];
                if (peek[0] === DIFF_INSERT) {
                    ins[1] += peek[1];
                }
                else if (peek[0] === DIFF_DELETE) {
                    del[1] += peek[1];
                }
                else {
                    const bound = findRightWordBound(peek, del, ins);
                    if (bound === -1) {
                        del[1] += peek[1];
                        ins[1] += peek[1];
                    }
                    else {
                        del[1] += peek[1].slice(0, bound);
                        ins[1] += peek[1].slice(0, bound);
                        diffs[i] = [peek[0], peek[1].slice(bound)];
                        break;
                    }
                }
                i++;
            }
            result.push(del, ins);
            start = null;
        }
        else if (cur[0] === DIFF_DELETE && !startsWithDelimiter(cur[1]) && start) {
            const nextStart = moveStartBound(start, result, cur);
            if (start !== nextStart) {
                result.push(cur);
                result.push([DIFF_INSERT, start[1].slice(nextStart[1].length)]);
            }
            else {
                result.push(cur);
            }
            start = null;
            i++;
        }
        else if (cur[0] === DIFF_INSERT && !startsWithDelimiter(cur[1]) && start) {
            const nextStart = moveStartBound(start, result, cur);
            if (start !== nextStart) {
                result.push([DIFF_DELETE, start[1].slice(nextStart[1].length)]);
                result.push(cur);
            }
            else {
                result.push(cur);
            }
            start = null;
            i++;
        }
        else {
            start = cur[0] === DIFF_EQUAL ? cur : null;
            result.push(cur);
            i++;
        }
    }
    return result;
}
function moveStartBound(start, result, ...tokens) {
    const nextStart = updateStartBound(start, ...tokens);
    if (nextStart !== start) {
        result.pop();
        if (nextStart[1] !== '') {
            // Edge case: bound update resulted in empty prefix,
            // do not add it
            result.push(nextStart);
        }
        return nextStart;
    }
    return start;
}
/**
 * Check if character at given location in string is a word delimiter
 */
function isWordDelimiter(ch) {
    return wordDelimiters.includes(ch);
}
/**
 * Finds location of word bound in given string. Returns `-1` if word bound wasn’t
 * found
 */
function findRightWordBound(eq, del, ins) {
    const text = eq[1];
    for (let i = 0; i < text.length; i++) {
        const ch = text.charCodeAt(i);
        if (numberPunct.includes(ch)) {
            // Possibly a decimal delimiter or number formatter
            if (isNumber(text.charCodeAt(i + 1)) || isNumber(text.charCodeAt(i - 1))) {
                continue;
            }
            if (i === 0 && (isNumber(lastCharCode(del[1])) || isNumber(lastCharCode(ins[1])))) {
                continue;
            }
        }
        if (isWordDelimiter(text.charCodeAt(i))) {
            return i;
        }
    }
    return -1;
}
/**
 * Updates given diffs so they point at the beginning of the word bound
 */
function updateStartBound(start, ...tokens) {
    const end = start[1].length - 1;
    let i = end;
    while (i >= 0) {
        const ch = start[1].charCodeAt(i);
        if (numberPunct.includes(ch) && isNumber(start[1].charCodeAt(i - 1))) {
            // A decimal delimiter or formatter character of number
            i -= 2;
            continue;
        }
        if (isWordDelimiter(ch)) {
            break;
        }
        i--;
    }
    if (i !== end) {
        const offset = end - i;
        tokens.forEach(t => t[1] = start[1].slice(-offset) + t[1]);
        return [start[0], start[1].slice(0, -offset)];
    }
    return start;
}
/**
 * Check if given text starts with word delimiter
 */
function startsWithDelimiter(text) {
    return isWordDelimiter(text.charCodeAt(0));
}
function lastCharCode(text) {
    return text.charCodeAt(text.length - 1);
}

/**
 * Calculates diff between given parsed document and produces new model with diff
 * tokens in it. This model can be restored into a final XML document
 */
function diff(from, to, options = createOptions()) {
    const dmp = new diff_match_patch();
    if (options.dmp) {
        Object.assign(dmp, options.dmp);
    }
    let diffs = dmp.diff_main(from.content, to.content);
    dmp.diff_cleanupSemantic(diffs);
    if (options.wordPatches) {
        diffs = updateWordBounds(diffs);
    }
    const toTokens = to.tokens.slice();
    let tokens = [];
    let offset = 0;
    let fromOffset = 0;
    diffs.forEach(d => {
        if (d[0] === DIFF_DELETE && d[1]) {
            // Removed fragment: just add deleted content to result
            tokens.push({
                name: d[1],
                type: 103 /* Delete */,
                location: offset,
                value: `<del>${reconstructDel(from, fromOffset, d[1], options.preserveTags)}</del>`
            });
            fromOffset += d[1].length;
        }
        else if (d[0] === DIFF_INSERT) {
            // Inserted fragment: should insert open and close tags at proper
            // positions and maintain valid XML nesting
            tokens = handleInsert(d[1], offset, toTokens, tokens);
            offset += d[1].length;
        }
        else if (d[0] === DIFF_EQUAL) {
            // Unmodified content
            offset += d[1].length;
            fromOffset += d[1].length;
            // Move all tokens of destination document to output result
            while (toTokens.length) {
                const first = toTokens[0];
                // if (first.location > offset || (first.location === offset && first.type === ElementType.Open)) {
                //     break;
                // }
                if (first.location > offset) {
                    break;
                }
                if (first.location === offset && first.type === 1 /* Open */) {
                    // Handle edge case. In the following examples:
                    // – aa <div>bb cc</div>
                    // – aa bb <div>cc</div>
                    // ...removing `dd ` results DELETE token with the same _text_
                    // location, yet in first case it should be outside `<div>` and
                    // in second case – inside `<div>`.
                    // In case if token touches the edge of open tag, we should detect
                    // if this token is inside or outside the same tag in `from`
                    // document
                    const { stack } = getElementStack(from, fromOffset);
                    let found = false;
                    while (stack.length) {
                        if (stack.pop().name === first.name) {
                            found = true;
                            break;
                        }
                    }
                    if (!found) {
                        break;
                    }
                }
                tokens.push(toTokens.shift());
            }
        }
    });
    if (options.compact) {
        tokens = compactIns(tokens, options);
        tokens = compactDel(to.content, tokens, options);
    }
    return {
        tokens: tokens.concat(toTokens),
        content: to.content
    };
}
/**
 * Applies INSERT patch operation
 * @param value Added content
 * @param pos Current output location
 * @param tokens Array of tag tokens *right* to given `pos` location.
 * Will be mutated: tokens will move into `output` array.
 * @param output Array of output tokens. Will be mutated with diff and tag tokens.
 * @return Returns `output` array
 */
function handleInsert(value, pos, tokens, output) {
    const end = pos + value.length;
    let tag;
    insOpen(output, pos);
    while (tokens.length && tokens[0].location < end) {
        tag = tokens.shift();
        if (isTagToken(tag)) {
            insClose(output, tag.location);
            output.push(tag);
            insOpen(output, tag.location);
        }
        else {
            output.push(tag);
        }
    }
    insClose(output, end);
    return output;
}
function insOpen(tokens, location) {
    tokens.push({ name: 'ins', type: 101 /* InsertBefore */, location, value: '<ins>' });
}
function insClose(tokens, location) {
    tokens.push({ name: 'ins', type: 102 /* InsertAfter */, location, value: '</ins>' });
}
/**
 * Optimizes given token list by removing meaningless insertion tags
 */
function compactIns(tokens, options) {
    const result = [];
    tokens = tokens.slice();
    while (tokens.length) {
        const token = tokens.shift();
        if (isType(token, 101 /* InsertBefore */)) {
            if (isType(tokens[0], 102 /* InsertAfter */) && token.location === tokens[0].location) {
                // Empty token
                tokens.shift();
                continue;
            }
            if (isType(tokens[0], 100 /* Space */) && isType(tokens[1], 102 /* InsertAfter */)) {
                // Inserted whitespace.
                // It can be either significant (`ab` -> `a b`) or insignificant,
                // if this whitespace is empty or is right after non-inline element
                const prev = result[result.length - 1];
                if (token.location === tokens[1].location
                    || (
                    // Check that only whitespace was inserted
                    (tokens[1].location - token.location) === tokens[0].offset
                        && isTagToken(prev) && !isInlineElement(prev, options))) {
                    result.push(tokens.shift());
                    tokens.shift();
                    continue;
                }
            }
        }
        result.push(token);
    }
    return result;
}
/**
 * Optimizes given token list by removing meaningless delete tokens
 */
function compactDel(content, tokens, options) {
    return tokens.filter((token, i) => {
        if (isType(token, 103 /* Delete */) && isWhitespace(token.name)) {
            if (isWhitespace(content.charAt(token.location - 1)) || isWhitespace(content.charAt(token.location + 1))) {
                // There’s whitespace before or after deleted whitespace token
                return false;
            }
            if (!isInlineElement(tokens[i - 1], options) && !isInlineElement(tokens[i + 1], options)) {
                // Whitespace between non-inline elements
                return false;
            }
        }
        return true;
    });
}
/**
 * Check if given string is whitespace-only
 */
function isWhitespace(str) {
    return str ? /^[\s\n\r]+$/.test(str) : false;
}
/**
 * Check if given token has specified `type`
 */
function isType(token, type) {
    return token ? token.type === type : false;
}
/**
 * Check if given token contains tag
 */
function isTagToken(token) {
    return token.type === 1 /* Open */
        || token.type === 2 /* Close */
        || token.type === 3 /* SelfClose */;
}
/**
 * Check if given token is an inline-level HTML element
 */
function isInlineElement(token, options) {
    return token && options.inlineElements.includes(token.name);
}
/**
 * Reconstructs deleted fragment of original documents with tag structure
 * of host document
 */
function reconstructDel(model, pos, value, tags) {
    if (tags.length === 0) {
        return value;
    }
    // tslint:disable-next-line: prefer-const
    let { stack, i } = getElementStack(model, pos);
    const endPos = pos + value.length;
    let result = '';
    // Check which elements should be kept in deleted fragment
    for (const token of stack) {
        if (tags.includes(token.name)) {
            result += token.value;
        }
    }
    // Walk up to the end of text fragment and check inner structure
    let offset = 0;
    while (i < model.tokens.length) {
        const token = model.tokens[i++];
        if (token.location > endPos) {
            break;
        }
        result += value.slice(offset, token.location - pos);
        offset = token.location - pos;
        if (token.type === 3 /* SelfClose */ && tags.includes(token.name)) {
            result += token.value;
        }
        else if (token.type === 1 /* Open */) {
            stack.push(token);
            if (tags.includes(token.name)) {
                result += token.value;
            }
        }
        else if (token.type === 2 /* Close */) {
            stack.pop();
            if (tags.includes(token.name)) {
                result += token.value;
            }
        }
    }
    // Close the remaining elements
    while (stack.length) {
        const token = stack.pop();
        if (tags.includes(token.name)) {
            result += token.value;
        }
    }
    return result + value.slice(offset);
}
/**
 * Collects element stack for given text location
 */
function getElementStack(model, pos) {
    const stack = [];
    let i = 0;
    while (i < model.tokens.length) {
        const token = model.tokens[i];
        if (token.location > pos) {
            break;
        }
        if (token.type === 1 /* Open */) {
            stack.push(token);
        }
        else if (token.type === 2 /* Close */) {
            stack.pop();
        }
        i++;
    }
    return { stack, i };
}

const reSpace = /[\s\r\n]+/g;
/**
 * Parses given XML source into internal model for diff
 */
function parse(text, options = createOptions()) {
    const { consume, finalize } = createConsumer(text, options);
    scan(text, consume, options);
    return finalize(text.length);
}
/**
 * Creates XML markup consumer for given text. Returns object with `consume` function
 * that should receive markup tokens and `finalize` function to produce final
 * document model
 * @param text Original source code being parsed
 */
function createConsumer(text, options = createOptions()) {
    const baseStart = options.baseStart || 0;
    const state = {
        text,
        tokens: [],
        options,
        content: '',
        hasContent: false,
        prev: baseStart,
        offset: 0
    };
    return {
        state,
        consume(name, type, start, end) {
            if (state.options.normalizeSpace) {
                normalizeWhitespace(state.prev, start, state);
            }
            else {
                state.content += text.slice(state.prev, start);
            }
            const token = {
                name, type,
                value: text.substring(start, end),
                location: start - baseStart - state.offset,
                order: start
            };
            state.tokens.push(token);
            state.prev = end;
            state.offset += token.value.length;
            if (shouldAddSpaceDelimiter(name, type, state)) {
                // Closing block-level element. In case if `wordPatches` option enabled,
                // we should add whitespace to prevent word collapsing.
                // Add empty whitespace token into document to suppress whitespace
                // added to document content
                state.tokens.push({
                    name: '#whitespace',
                    type: 100 /* Space */,
                    value: '',
                    location: token.location,
                    offset: 1,
                    order: token.location
                });
                state.content += ' ';
                state.hasContent = false;
                state.offset--;
            }
        },
        finalize(baseEnd = text.length) {
            if (state.options.normalizeSpace) {
                normalizeWhitespace(state.prev, baseEnd, state);
                fixTrailingSpace(state);
            }
            else {
                state.content += text.slice(state.prev, baseEnd);
            }
            return {
                tokens: state.tokens,
                content: state.content
            };
        }
    };
}
/**
 * Normalizes whitespace characters: replaces all whitespace occurrences
 * (including adjacent) with a single space and returns tokens with original space
 * characters
 */
function normalizeWhitespace(from, to, state) {
    let m;
    let prev = 0;
    let hasSpace = false;
    const fragment = state.text.substring(from, to);
    while (m = reSpace.exec(fragment)) {
        hasSpace = true;
        if (m[0] !== ' ' || (m.index === 0 && !state.hasContent)) {
            // Found formatted whitespace, collapse it to a single space or remove
            // if it’s preceded by another whitespace
            state.content += fragment.slice(prev, m.index);
            const token = {
                name: '#whitespace',
                type: 100 /* Space */,
                value: m[0],
                location: from + m.index - state.offset - (state.options.baseStart || 0),
                order: from + m.index
            };
            if (state.hasContent) {
                // Not a trailing (significant) whitespace, replace it with single space
                state.content += ' ';
                token.offset = 1;
                state.offset--;
            }
            prev = m.index + m[0].length;
            state.tokens.push(token);
            state.offset += token.value.length;
            state.hasContent = prev < fragment.length;
        }
        else {
            // Entered trailing space?
            state.hasContent = m.index + m[0].length < fragment.length;
        }
    }
    if (fragment && !hasSpace) {
        state.hasContent = true;
    }
    state.content += fragment.slice(prev);
}
/**
 * Fixes trailing space in accumulated content
 */
function fixTrailingSpace(state) {
    const trailing = state.content[state.content.length - 1] === ' ';
    if (trailing) {
        // Offset trailing tokens with new space
        for (let i = state.tokens.length - 1; i >= 0; i--) {
            const token = state.tokens[i];
            if (token.type === 100 /* Space */) {
                token.offset = 0;
                break;
            }
            else {
                token.location--;
            }
        }
        state.content = state.content.slice(0, -1);
        state.hasContent = state.content.length > 0;
    }
}
/**
 * Check if extra space delimiter should be added for the given XML token.
 * Used to determine if a space character should be added after closing tag in order
 * to prevent word collapsing in diff
 */
function shouldAddSpaceDelimiter(name, type, state) {
    return Boolean(state.options.wordPatches
        && (type === 2 /* Close */ || type === 3 /* SelfClose */)
        && !state.options.inlineElements.includes(name)
        && state.content
        && state.content.slice(-1) !== ' ');
}

/**
 * Calculates diff between two XML documents and returns `to` document content
 * with diff patches applied to it as `<ins>`/`<del>` tags
 */
function diffDocuments(from, to, options) {
    const opt = createOptions(options);
    const fromDoc = parse(from, opt);
    const toDoc = parse(to, opt);
    const diffDoc = diff(fromDoc, toDoc, opt);
    return stringify(diffDoc);
}
/**
 * Returns restored source from given parsed model
 */
function stringify(model) {
    let offset = 0;
    return model.tokens.map(token => {
        const result = model.content.slice(offset, token.location) + token.value;
        offset = token.location + (token.offset || 0);
        return result;
    }).join('') + model.content.slice(offset);
}

export default diffDocuments;
export { createConsumer, createOptions, diff, parse, stringify };
//# sourceMappingURL=xml-diff.es.js.map
